"""
Pipeline celery signal handlers

"""
from copy import deepcopy

from celery.signals import task_postrun

from pipeline.actions import PipelineTask
from pipeline.eval import safe_eval

import logging
logger = logging.getLogger(__name__)



@task_postrun.connect()
def task_postrun_handler(*args, **kwargs):
    """Action post-run signal handler to execute handlers.
    """
    assert 'retval' in kwargs
    assert 'task' in kwargs
    assert 'sender' in kwargs
    assert 'state' in kwargs

    # skip failed tasks
    if kwargs['state'] != 'SUCCESS':
        return

    # skip tasks that aren't part of pipeline
    if not isinstance(kwargs['sender'], PipelineTask):
        return

    logger.debug('postrun handler triggered {} {}'.format(args, kwargs))

    task_kwargs = deepcopy(kwargs['kwargs'])  # inception!
    state = task_kwargs.get('_pipeline_chain_state', {})
    if 'children' in state:
        child_tasks = state['children']
        #TODO: check to make sure children never reaches beyond a single action
        del state['children']  # may not be necessary

        context = task_kwargs['_pipeline_chain_state']['build_context']

        for action, predicate in child_tasks:
            #TODO: execute children in parallel

            try:
                should_execute = safe_eval(
                    predicate,
                    context
                )
            except:
                should_execute = False

            if should_execute:
                action.prepare(context)
                action.partial.delay()



""" Jenkins API

TODO:
    Mayneed to support multiple sources.
    We are using inheritance,we may need to switch to composition

THIS DOES NOT WORK.  OLD CODE.
This module is not registered in pipeline.__init__.
"""
from urllib.request import urlopen, Request, build_opener, ProxyHandler, install_opener
import urllib
import json
from urllib.error import HTTPError
import base64
from jenkinsapi.api import get_artifacts
from jenkinsapi.jenkins import Jenkins
try:
    from urllib.parse import urlparse
except ImportError:
    import urlparse

from pipeline.actions import action, TaskResult
from pipeline.bases import Source
from pipeline.modules.github.pullrequest import PullRequest
from pipeline.tasks import TaskResult
import logging
logger = logging.getLogger(__name__)
from pipeline.command import CommandResult

class ArtifactTaskResult(TaskResult):
    # this is getting a little ridiculous.
    #need to replace 'Source' with a writable Context
    def __init__(self, result, artifact_url, user, passwd):
        super(ArtifactTaskResult, self).__init__(result)
        self.url = artifact_url
        self.user = user
        self.passwd = passwd


@action(name='jenkins-artifact')
def jenkins_artifact(self, retval, source, **kwargs):
    """Pull an artifact from jenkins.
    """
    def auth_headers(username, password):
        s = '{}:{}'.format(username, password)
        b = bytes(s, 'ascii')
        return 'Basic {}'.format(
            base64.b64encode(b).decode('ascii')
        )

    artifact_file = kwargs.pop('artifact')

    parsed = urlparse(source.target_url)
    jenkins = "{}://{}".format(
        parsed.scheme, parsed.netloc
    )
    parts = parsed.path.split('/')

    job = parts[2]
    build = parts[3]

    prxy = ProxyHandler({})
    opener = build_opener(prxy)
    install_opener(opener)
    #TODO move to env var
    auth = auth_headers('E001882', '567aca75e52ccdf8f4feaf5f367c7c04')
    api_url = '{}/job/{}/{}/api/json'.format(jenkins, job, build)
    req = Request(api_url)
    req.add_header('Authorization', auth)
    try:
        response =  urlopen(req).read()
    except HTTPError as ex:
        logger.critical(ex)
        return TaskResult(False)

    data = json.loads(response.decode('utf-8'))
    artifacts = data['artifacts']
    for artifact in artifacts:
        if artifact['fileName'] == artifact_file:
            artifact_url = '{}/job/{}/{}/artifact/{}'.format(jenkins, job, build, artifact['relativePath'])
            #req = Request(artifact_url)
            #req.add_header('Authorization', auth)
            #try:
            #    response =  urlopen(req).read()
            #except HTTPError as ex:
            #    logger.critical(ex)
            #    return TaskResult(False)
            
            return ArtifactTaskResult(True, artifact_url, 'E001882', '567aca75e52ccdf8f4feaf5f367c7c04')

    logger.error('Didnt find artifact {} at {}'.format(artifact_file, api_url))
    return TaskResult(False)


class PrBuilderStatusSource(PullRequest):
    __id = 'pr_status'
    
    def __init__(self, build_url, *args, **kwargs):
        self.target_url = build_url
        super(PrBuilderStatusSource, self).__init__(*args, **kwargs)

#import time
#import urllib
#import uuid
#import simplejson as json
#from simplejson.decoder import JSONDecodeError
#import logging
##from enum import Enum
#import requests
#from requests.exceptions import HTTPError
#from datetime import timedelta
#
#from django.conf import settings
#
#
##logging.basicConfig()
#log = logging.getLogger(__name__)
#
#class BuildError(Exception):
#    pass
#
#class Jenkins(object):
#    def __init__(self, host, user, token):
#        self.host = host
#        self.user = user
#        self.token = token
#
#class ParameterizedJob(object):
#    """
#    deals with https://issues.jenkins-ci.org/browse/JENKINS-13546
#        - requires the job contain an 'ident' param do we can
#        figure out it's build number from the queue.
#    """
#    def __init__(self, jenkins, job_name):
#        self.jenkins = jenkins
#        self.job_name = job_name
#        self.ident = uuid.uuid4()  # needed to find the jobid
#        self.queue__id = None
#        self.build_number = None
#        self.session = requests.Session()
#
#    def build(self, **job_params):
#
#        params = job_params.copy()
#        params.update({
#            'token': self.jenkins.token,
#            'delay': settings.BUILD_SUBMIT_DELAY,
#            'ident': self.ident
#        })
#        url = "{0}/job/{1}/buildWithParameters?{2}".format(
#            self.jenkins.host, self.job_name, urllib.urlencode(params)
#        )
#        log.debug("building jenkins job at %s" % url)
#
#        try:
#            # enqueue the job
#            resp = self.session.get(url, auth=(self.jenkins.user, self.jenkins.token))
#            resp.raise_for_status()
#        except HTTPError as e:
#            log.error("error queueing jenkins job %s" % str(e))
#        else:
#            # find our job in the queue
#            self.queue__id = self.get_queue__id()
#            if self.queue__id is None:
#                raise BuildError("Couldn't get queue id")
#
#            # get our build number
#            self.build_number = self.poll_for_build_number()
#            if self.build_number is None:
#                raise BuildError("Couldn't find job number")
#
#    def get_queue_id(self):
#        """Find a job the job queue, using an artificial `ident` parameter that
#        was passed into the build.
#        This is required until Jenkins bug https://issues.jenkins-ci.org/browse/JENKINS-13546
#        is resolved.
#        """
#        log.debug('attempting to get queue id for our job %s' % self.job_name)
#        try:
#            resp = self.session.get("%s/queue/api/json" % self.jenkins.host)
#            resp.raise_for_status()
#            queue = json.loads(resp.content)
#        except HTTPError as e:
#            log.error("http error getting queue id %s" % str(e))
#        except JSONDecodeError:
#            log.error("erorr decoding response")
#        else:
#            for item in queue['items']:
#                if not item['task']['name'] == self.job_name:
#                    continue
#                params = [a['parameters'] for a in item['actions'] if 'parameters' in a][0]
#                ident_param = [i['value'] for i in params if i['name'] == 'ident'][0]
#                if ident_param == unicode(self.ident):
#                    log.debug("found queue id for job")
#                    return str(item['id'])
#
#    def poll_for_build_number(self):
#        """Find a job's build id in the queue details.
#        The build ID is not immediately available when the job is queued.
#
#        Make a queue api request each ``QUEUE_POLL_WAIT`` seconds until a response
#        containing build number is received, or ``QUEUE_POLL_MAX_WAIT`` seconds is reached.
#        """
#        max_wait = settings.QUEUE_POLL_MAX_WAIT
#        while True:
#            try:
#                log.debug('attempting to get build number for queue item %s' % self.queue_id)
#                response = self.session.get(
#                    "%s/queue/item/%s/api/json" % (self.jenkins.host, self.queue_id)
#                )
#                response.raise_for_status()
#                queue_info = json.loads(response.content)
#            except HTTPError as e:
#                log.warning("could not get the jenkins build number %s; bailing" % str(e))
#                break
#            except JSONDecodeError:
#                log.warning("decode error")
#                break
#            else:
#                if 'executable' in queue_info and 'number' in queue_info['executable']:
#                    build_number = str(queue_info['executable']['number'])
#                    log.debug("got build number %s" % build_number)
#                    return build_number
#
#                time.sleep(settings.QUEUE_POLL_WAIT)
#                max_wait -= settings.QUEUE_POLL_WAIT
#                if max_wait < 0:
#                    log.error("max poll limit reached (%s seconds) for queue item %s!" % (
#                        settings.QUEUE_POLL_MAX_WAIT, self.queue_id)
#                    )
#                    break
#
#    def poll_for_completion(self):
#        """Wait for a build to run to completion.
#
#        Make a build api request each ``BUILD_POLL_WAIT`` seconds until a response
#        indicating completion is received, or ``BUILD_POLL_MAX_WAIT`` seconds is reached.
#        """
#        max_wait = settings.BUILD_POLL_MAX_WAIT
#        log.debug('polling build %s, waiting for completion' % self.build_number)
#        while True:
#            try:
#                response = self.session.get(
#                    "%s/job/%s/%s/api/json" % (self.jenkins.host, self.job_name, self.build_number)
#                )
#                response_data = json.loads(response.content)
#            except HTTPError as e:
#                log.error("polling error %s - bailing" % str(e))
#                raise BuildError("http error while polling %s" % str(e))
#            except JSONDecodeError:
#                log.error("Error decoding")
#                raise BuildError("decode error")
#            else:
#                result = BuildResult.from_blob(response_data)
#
#                if not result.building:
#                    log.debug("job has completed")
#                    return result
#
#                log.debug("build %s still running; waiting %s sec" % (
#                    self.build_number, settings.BUILD_POLL_WAIT)
#                )
#                time.sleep(settings.BUILD_POLL_WAIT)
#
#                max_wait -= settings.BUILD_POLL_WAIT
#                if max_wait < 0:
#                    log.error("max poll limit reached (%s seconds) for build %s!" % (
#                        settings.BUILD_POLL_MAX_WAIT, self.build_number)
#                    )
#                    raise BuildError("polled too long")
#
#    @property
#    def url(self):
#        return "%s/job/%s/%s/" % (self.jenkins.host, self.job_name, self.build_number)
#
#    @property
#    def console_url(self):
#        return "%s/job/%s/%s/console" % (self.jenkins.host, self.job_name, self.build_number)
#
#    def __str__(self):
#        return vars(self.__dict__)
#
##
## class JobStatus(Enum):
##     building = 1
##     aborted = 2
##     succeeded = 3
##     failed = 4
##     unstable = 5
##
## job_status = {
##     None: JobStatus.building,
##     "ABORTED": JobStatus.aborted,
##     "SUCCESS": JobStatus.succeeded,
##     "FAILURE": JobStatus.failed,
##     "UNSTABLE": JobStatus.unstable
## }
#
#class BuildResult(object):
#    """Encapsulate results of build."""
#    def __init__(self):
#        self.status = None
#        self.duration = None
#
#    @classmethod
#    def from_blob(cls, blob):
#        ji = BuildResult()
#        if bool(blob['building']):
#            ji.status = JobStatus.building
#        else:
#            ji.status = job_status[blob['result']]
#        ji.duration = timedelta(milliseconds=blob['duration'])
#        return ji
#
#    @property
#    def building(self):
#        return self.status == JobStatus.building
#
#    def __str__(self):
#        return "Status: %s, duration: %s" % (repr(self.status), self.duration)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pipeline', '0002_configuration_app_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='configuration',
            name='app_name',
            field=models.CharField(max_length=64, blank=True, null=True),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pipeline', '0004_auto_20150530_1415'),
    ]

    operations = [
        migrations.RenameField(
            model_name='configurationsource',
            old_name='source_class',
            new_name='event',
        ),
        migrations.RemoveField(
            model_name='configuration',
            name='source',
        ),
        migrations.AddField(
            model_name='configuration',
            name='criteria',
            field=jsonfield.fields.JSONField(default=[]),
        ),
        migrations.AddField(
            model_name='configuration',
            name='events',
            field=jsonfield.fields.JSONField(default=[]),
        ),
    ]



import time

from pipeline.actions import action, TaskResult
from pipeline.command import BakedCommand, register_baked_command_task

import logging
logger = logging.getLogger(__name__)

class EchoCommand(BakedCommand):
    """This command will echo whatever the
    souurce was."""
    __id = 'echo_test_command'
    def instructions(self):
        return [['echo', self.context.source]]

register_baked_command_task('echo_test_command')

class DummyFileSource(object):
    def __init__(self, filename, contents):
        self.acquisition_instructions = {
            'command': 'echo {} >> {}'.format(
                contents, filename
            )
        }
        
@action
def print_something(self, source, the_thing=None):

    logger.error('printing something')
    for i in range(3):
        logger.error('{} {}'.format(the_thing, i))
        time.sleep(0.5)
    logger.error('done printing something')

    return TaskResult(False, output=the_thing)

@action
def log_stuff(self, source):
    """Print some stuff to the console."""
    logger.error(source)



@action
def dummy_action(self, source, **kwargs):
    return 1




import os
import pwd
import six
import shutil
import logging                  # <-- beautiful
import tempfile                  # <-- imports
import platform                   # <-- inspired by
import functools                   # <-- miki725!
import subprocess
import contextlib
from copy import deepcopy

#TODO fix commandsession library import
from commandsession.commandsession import CommandSession

from django.conf import settings
from pipeline.bases import Registry

from pipeline.utils import rand_suffix


logger = logging.getLogger(__name__)


def get_current_user():
    """Return the current user.
    There are various ways to do this, and I've tried them all.
    """
    return pwd.getpwuid(os.getuid())[0]


def get_user_environment(username):
    """Get a user's environment as a dictionary.

    Requires that the current user has sudo privileges.

    :param username: a valid system username
    :returns: dict of env vars
    """
    if get_current_user() == username:
        cmd = 'env'
    else:
        cmd = 'sudo su -l {} -c env'.format(username)

    try:
        env_str = subprocess.check_output(
            cmd, stderr=subprocess.PIPE, shell=True
        )
    except subprocess.CalledProcessError as e:
        logger.error("Error getting user {}'s environment: {} {} {}".format(
            username, e.cmd, e.returncode, e.output.decode('utf-8'))
        )
        raise
    else:
        # `check_output` returns bytestring in py3
        # this is a nop in py2
        env_str = env_str.decode('utf-8')
        env = {
            l[0]:l[2] for l in
            [i.partition('=') for i in env_str.split()]
        }

        if hasattr(settings, 'PIPELINE_INJECT_ENVIRON'):
            env.update(settings.PIPELINE_INJECT_ENVIRON)

        return env


class WorkspaceError(Exception):
    pass


class Workspace(metaclass=Registry):
    """Base class for Workspaces."""


class workspace(contextlib.ContextDecorator, Workspace):
    """Workspace.

    A container for interacting with the system on a worker.

    Creates/deletes on-disk working directory and handles
    execution of shell commands within the correct environment.

    Must be instantiated on the worker.
    """
    __id = 'workspace'

    def __init__(self, name=None, basepath=None, hints=None, delete=True, reusable=False, session=None):
        """`reusable` param is only used for testig right now.

        :param name:
            combined with ``hints`` and the name of this class,
            determines the on-disk directory name for this workspace.
        :param basepath: location of temp area for creating workspace
        :param hints: list, see ``name``
        :param delete: boolean, determines whether to delete this workspace
            after use
        :param reusable: boolean, if False the on-disk directory has a
            random value prepended to it.
        :param session: CommandSession instance, if provided the workspace will
            use the given session, otherwise it will create one.  Useful
            if the cller wants to inspect the session log after-thefact.
        """
        self.user = get_current_user()
        self.environ = get_user_environment(self.user)

        if not basepath:
            basepath = settings.PIPELINE_WORKSPACE_ROOT

        # haaaaaaaaaaaaaack for OSX w/ brew
        if platform.uname()[0] == 'Darwin':
            self.environ['PATH'] = "{}:{}".format(
                '/usr/local/bin',
                self.environ['PATH']
            )

        logger.debug('initializing workspace for user {} in {}'.format(
            self.user, basepath
        ))

        self.delete = delete

        path_parts = hints or []
        assert isinstance(path_parts, list)

        if not reusable:
            path_parts.append(rand_suffix())

        prefix = [x for x in [self.__class__.__name__.lower(), name] if x]
        pathstr = '-'.join([str(x) for x in prefix + path_parts])

        self.location = self.cwd = os.path.join(
            basepath or tempfile.gettempdir(), pathstr
        )

        self.session = session or CommandSession(
            stream=True,
            env=self.environ,
            cwd=self.cwd
        )

    def acquire_source(self, source):
        """Run a source's acquisition instructions in the workspace.
        """
        if not hasattr(source, 'acquisition_instructions'):
            logger.debug('source {} does not need to be acquired'.format(source))
            return 0

        for item in source.acquisition_instructions:
            self.session.check_call(item)


    def __enter__(self):
        """Populate the workspace.
        """
        assert not os.path.exists(self.location)
        logging.debug('creating workspace {}'.format(self.location))
        try:
            os.mkdir(self.location)
        except OSError as exc:
            logger.error("Could not create workspace: %s" % str(exc))
            raise WorkspaceError(
                "Could not create workspace at %s" % self.location
            ) from exc

        return self

    def __exit__(self, *exc):
        # this is fricking dangerous.  figure something out.
        logging.debug('deleting workspace {}'.format(self.location))
        if self.delete:
            shutil.rmtree(self.location)

        return False


class python_workspace(workspace):
    """Python2-specific workspace.
    Wraps execution in a virtualenv, by prepending env directory to PATH.
    """
    __id = 'python_workspace'
    venv = 'virtualenv'

    def __enter__(self):
        super(python_workspace, self).__enter__()
        ret = self.session.check_call([self.venv, 'env'])
        if not ret == 0:
            raise WorkspaceError('Could not create virtualenv')

        self.environ['PATH'] = "{}:{}".format(
            os.path.join(self.location, 'env', 'bin'),
            self.environ['PATH']
        )
        self.environ['VIRTUAL_ENV'] = os.path.join(self.location, 'env')

        return self


class python3_workspace(python_workspace):
    """Python3-specific workspace.
    Wraps execution in a virtualenv, by prepending env directory to PATH.
    """
    __id = 'python3_workspace'
    venv = 'pyvenv'


def get_workspace(workspace_type, *args, **kwargs):
    """Temporary factory function."""
    if workspace_type in ('python', 'python2'):
        return python_workspace(*args, **kwargs)
    elif workspace_type == 'python3':
        return python3_workspace(*args, **kwargs)
    return workspace(*args, **kwargs)


class v1_workspace(contextlib.ContextDecorator, Workspace):
    """DEPRECATED

    Workspace.

    A container for interacting with the system on a worker.

    Creates/deletes on-disk working directory and handles
    execution of shell commands within the correct environment.

    TODO:
        - implement workspace reusability. will allow us to leverage
        data locality, to execute commands in a pre-existing workspace
        eliminating startup costs
        - allow caller to specify username?
        - doesnt run bash with startup file etc..
        - file descriptors for stdout and stderr so that entire command
        output can be propagated to other tasks
        - experiment with various methods of subprocess.  currently using call(),
        which is simple, but it might not be enough.  for instance, capturing
        error and returning to caller, instead of diaplying to stderr of process

    """
    __id = 'v1_workspace'

    def __init__(self, name=None, basepath=None, hints=None, delete=True, reusable=False):
        """`reusable` param is only used for testig right now.

        :param name:
            combined with ``hints`` and the name of this class,
            determines the on-disk directory name for this workspace.
        :param basepath: location of temp area for creating workspace
        :param hints: list, see ``name``
        :param delete: boolean, determines whether to delete this workspace
            after use
        :param reusable: boolean, if False the on-disk directory has a
            random value prepended to it.

        """
        self.user = get_current_user()
        self.environ = get_user_environment(self.user)

        if not basepath:
            basepath = settings.PIPELINE_WORKSPACE_ROOT

        # haaaaaaaaaaaaaack for OSX w/ brew
        if platform.uname()[0] == 'Darwin':
            self.environ['PATH'] = "{}:{}".format(
                '/usr/local/bin',
                self.environ['PATH']
            )

        logger.debug('initializing workspace for user {} in {}'.format(
            self.user, basepath
        ))

        self.delete = delete

        path_parts = hints or []
        assert isinstance(path_parts, list)

        if not reusable:
            path_parts.append(rand_suffix())

        prefix = [x for x in [self.__class__.__name__.lower(), name] if x]
        pathstr = '-'.join([str(x) for x in prefix + path_parts])

        self.location = self.cwd = os.path.join(
            basepath or tempfile.gettempdir(), pathstr
        )

    def acquire_source(self, source):
        """Run a source's acquisition instructions in the workspace.
        If the source specifies that the installation resulted in
        a change of directory, alter self.cwd to match.

        #TODO use some abstraction for the instructions, instead of a dict
        """
        if not hasattr(source, 'acquisition_instructions'):
            logger.debug('source {} does not need to be acquired'.format(source))
            return 0

        instructions = source.acquisition_instructions

        # run install command
        # FIXME SOON
        if isinstance(instructions['command'], six.string_types):
            # hack for debugging
            import subprocess
            ret = subprocess.call(
                instructions['command'],
                shell=True
            )
        else:
            ret = self.run_command(
                instructions['command'],
            )

        if not ret == 0:
            logger.error("couldnt run command {} for source {}".format(
                instructions['command'], source
            ))
            return ret

        # update workspace location with source subdir.
        # this may not work, good enough for now.
        if instructions['directory']:
            self.cwd = os.path.join(
                self.location, instructions['directory']
            )
            logger.debug('cwd changed to {}'.format(self.cwd))

        # run post-install commands, if any
        for command in instructions['post_commands']:
            ret = self.run_command(command)
            if not ret == 0:
                logger.error("couldnt run command for source {}".format(
                    command, source
                ))
                return ret

        return 0

    def __enter__(self):
        """Populate the workspace.
        """
        assert not os.path.exists(self.location)
        logging.debug('creating workspace {}'.format(self.location))
        try:
            os.mkdir(self.location)
        except OSError as exc:
            logger.error("Could not create workspace: %s" % str(exc))
            raise WorkspaceError(
                "Could not create workspace at %s" % self.location
            ) from exc

        return self

    def __exit__(self, *exc):
        # this is fricking dangerous.  figure something out.
        logging.debug('deleting workspace {}'.format(self.location))
        if self.delete:
            shutil.rmtree(self.location)

        return False

    def wrap_command(self, command, env=None):
        """Get a partial for running an arbitrary command.
        TODO: remove this. his was useful when we passed
        the callables into the celery task to micromanage them.
        """
        #assert isinstance(command, list), "not supporting subprocess shell=True"

        environ = deepcopy(self.environ)
        if env:
            environ.update(env)
        if isinstance(command, six.string_types):
            return functools.partial(
                subprocess.call,
                command, cwd=self.cwd, env=environ, shell=True
            )
        else:
            return functools.partial(
                subprocess.call,
                command, cwd=self.cwd, env=environ
            )

    def run_command(self, command, env=None):
        """Run a wrapped command."""
        logger.debug('running command {}'.format(command))
        command = self.wrap_command(command, env)
        #FIXME
        try:
            result = command()
        except Exception as e:
            logger.error(str(e))
            logger.error(self.environ)
            raise
        else:
            return result

    def get_output(self, command, env=None):
        """Temporary."""
        def fmt_output(what):
            return what.decode('utf-8')

        #assert isinstance(command, list), "not supporting subprocess shell=True"
        environ = deepcopy(self.environ)
        if env:
            environ.update(env)
        try:
            if isinstance(command, six.string_types):
                ret = subprocess.check_output(
                    command, cwd=self.cwd, env=environ, shell=True
                )
            else:
                ret = subprocess.check_output(
                    command, cwd=self.cwd, env=environ
                )
        except subprocess.CalledProcessError as exc:
            return fmt_output(exc.output), exc.returncode

        return fmt_output(ret), 0





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Configuration',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('actions', jsonfield.fields.JSONField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ConfigurationSource',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, auto_created=True, verbose_name='ID')),
                ('source_class', models.CharField(max_length=255)),
                ('criteria', jsonfield.fields.JSONField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='configuration',
            name='source',
            field=models.ForeignKey(to='pipeline.ConfigurationSource'),
            preserve_default=True,
        ),
    ]




class Registry(type):
    """Simple implementation of the Registry pattern using the
    declarative style.

    Classes that wish to implement a registry should utilize this as
    their metaclass.

    Subclasses of registries ("registered classes" should define the
    class attribute `__id`, which will serve as their key in the registry.

    Subclasses of registries that do not define an `__id` are considered
    'abstract', and will not be added to the registry.
    Their purpose should only be to serve to facilitate OOP patterns.
    """
    def __init__(cls, name, bases, nmspc):
        super(Registry, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, '_registry'):
            # assume base type here (superclass), which has no instances,
            # only implementations.
            cls._registry = {}
            return
        _id = '_{}__id'.format(name)
        _abstract = '_{}__abstract'.format(name)
        if hasattr(cls, _id):  # and not hasattr(cls, _abstract):
            cls._registry[getattr(cls, _id)] = cls  # noqa
            pass

    def factory(cls, name, *args, **kwargs):
        """Return an instance of the registered Thing identified by ``name``."""
        klass = cls.get(name)
        if not klass:
            raise KeyError('{} having name {} not found'.format(cls, name))
        return klass(*args, **kwargs)

    def get(cls, name):
        """Return a registred subclass by it's __id."""
        return cls._registry.get(name)

    def getlist(cls):
        """Return all the subclasses registered for a given class.
        """
        return filter(
            lambda k: issubclass(k, cls),
            cls._registry.values()
        )

    def find(cls, attr, value):
        """Find registered subclasses that define the attribute `attr`
        which has the value ``value``.
        """
        available = cls.getlist()
        found = []
        for v in available:
            if hasattr(v, attr) and getattr(v, attr) == value:
                found.append(v)
        return found



""" Pipeline Actions

Provide infrastructure for creating and scheduling celery tasks.
Action is a container for a task, to provide some syntactic
sugar without messing around with Task class internals.

To register actions, define a task function and a TaskAction subclass that
refers to it, or (preferably) use the @action decorator on the task function
itself.
"""

import logging
import six

from celery import shared_task, Task

from pipeline.workspace import get_workspace

logger = logging.getLogger(__name__)

class ActionRegistry(type):
    """Registry for action storage."""
    def __init__(cls, name, bases, nmspc):
        super(ActionRegistry, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, '_registry'):
            # top-level parent class, do not map to a task
            cls._registry = {}
            return
        if not hasattr(cls, '_name') or not hasattr(cls, '_task'):
            logger.error('cannot register cls {}'.format(cls))
            return
        cls._registry[cls._name] = cls._task  # noqa

    def get(cls, name):
        """Get a registered task function for a given name."""
        try:
            task_func = cls._registry[name]
        except KeyError:
            raise ValueError('{} not found in registry for {}'.format(
                name, cls.__name__
            ))
        else:
            return task_func


class TaskAction(metaclass=ActionRegistry):
    """Base Action class.  Subclasses need to define both
    ``_task`` and ``_name`` class attributes in order to be registered.
    We distinguish between 'task' and 'partial', even though 'partial'
    is derived from 'task'.  Having original task laying around is really
    helpful for testing, as you can make assertions using it that you cannot
    make with a partial.

    TODO: something is wonky here.  TaskAction instances arent really in
    the registry, the tasks underneath them are.
    For instance, I cannot run `TaskAction.get('some_name') and get a TaskAction
    back; what I get is an actual celery task. (the thing that will be in
    `self.task` after `TaskAction.__init__` is run). #FIXME
    Instead, i have to create a TaskAction instance providing the name of
    the task function.

    """

    def __init__(self, task_name, name=None, children=None, workspace=None, workspace_kwargs=None, **task_kwargs):
        self.task = self.__class__.get(task_name)
        self.partial = None
        self.name = name or "{}.{}".format(self.task.__module__, task_name)
        self.children = children or []
        self.workspace = workspace
        self.workspace_kwargs = workspace_kwargs or {}
        self.task_kwargs = task_kwargs or {}

    @property
    def prepared(self):
        """Executor shouldnt run tasks that havent been signatured"""
        return bool(self.partial)

    # def delay(self, *args, **kwargs):
    #     # used for testing; in the real world, we run actions
    #     # using an executor
    #     if self.partial:
    #         return self.partial.delay(*args, **kwargs)
    #     elif self.task:
    #         return self.task.delay(*args, **kwargs)
    #     raise Exception('not ready')  # TODO

    def prepare(self, source, build_context=None):  # **task_kwargs):
        """Return a task ready for a signature.
        Always use a bound task.
        All task_kwargs will be used to generate the partial.

        :param name: the action name in registry
        :returns: a TaskAction containing the task and partial
        """
        task = self.task
        kwargs = {}
        kwargs.update(self.task_kwargs)
        kwargs.update({
            '_pipeline_chain_state': {
                'children': self.children,
                'action_name': self.name,
            }
         })

        # This is normally done in ``pipeline_task_wrapper``, but that
        # is executed during task runtime on the worker.
        # Here we are running on the broker, priming the chain with
        # an initial build context.
        # TODO this responsibility probably belongs in the executor, not here.
        if build_context:
            kwargs['_pipeline_chain_state'].update(
                {'build_context': build_context}
            )

        kwargs.update({'_pipeline_workspace': {
            'klass': self.workspace,
            'kwargs': self.workspace_kwargs
        }})

        partial = task.s(
            source,
            **kwargs
        )

        #TODO: just return the partial instead of hanging off self
        # this will necessitate removal of delay() method, which I think
        # is only used for tests
        # ^^ This will require some plumbing changes - @mike
        #self.partial = partial
        return partial

    @classmethod
    def from_dict(cls, dct):
        """Build action from dictionary.
        Eventually we will pull things out of the dict and add
        them to the model.


        """
        children = []  # list of tuples
        if 'children' in dct:
            for item in dct['children']:
                # because predicate is not a normal task attribute, pop it before serializing
                # default of 'parent' will execute the __bool__ method of the parent's return
                # value, which should be a TaskResult or a subclass.
                predicate = item.pop('predicate', 'parent')
                children.append((
                    cls.from_dict(item),
                    predicate
                ))

        workspace_cls = dct.get('workspace')
        workspace_kwargs = dct.get('workspace_kwargs')

        task_action = cls(
            dct['action'],
            name=dct.get('name'),
            children=children,
            workspace=workspace_cls,
            workspace_kwargs=workspace_kwargs,
            **dct.get('kwargs', {})
        )

        return task_action


def register_action(name, task):
    """Convenience function to register a task/action relationship,
    without having to define a TaskAction subclass.
    """
    type('__TempCls', (TaskAction,), {'_name': name, '_task': task})

def pipeline_task_wrapper(f):
    """Small wrapper around celery tasks registered as Actions.
    Updates the build context supplied by pipeline with the results
    of the wrapped task, and then returns that context, which will
    appear, to the worker, to be the return value of the task.
    Coupled with mutable task signatures (a-la ``subtask()``), this
    allows the up-to-date build context to be passed from task to task
    in a celery chain like a truckstop hoo-a.
    """
    def newf(*args, **kwargs):

        assert len(args) > 0, 'task must be bound'  # missing 'self'
        assert isinstance(args[0], Task), 'task must be bound' # `self` must be a Task subclass

        self = args[0]
        source = args[1]

        args, kwargs = self.build_context.render_params(source, *args, **kwargs)
        ret = f(*args, **kwargs)

        # i think this scenario is only relevant for tests,
        #maybe remove it??  TODO
        if not hasattr(self, '_pipeline_chain_state'):
            # not a chained pipeline task
            raise NotImplementedError

        return self._pipeline_chain_state['build_context'].update(
            self._pipeline_chain_state['action_name'],
            ret
        )

    # There are some incompatibilities between celery and functools.wraps
    # with respect to sending wrapped functions as tasks.
    # Because celery relies on __module__.__name__ for task identification, and since
    # we don't want all tasks to have the same name 'pipeline.actions.newf', we must
    # manually modify __name__ and __module__.
    newf.__name__ = f.__name__
    newf.__module__ = f.__module__
    return newf


def action(*args, **kwargs):
    """Register an Action in pipeline.
    An action is a blahblah blah.
    args and kwargs are routed to ``celery.shared_task``,
    and may be used by pipeline as well (see `name``).
    """
    app_wrapper = shared_task
    def wrapper(**_kwargs):
        def __inner(func):
            tsk = app_wrapper(pipeline_task_wrapper(func), **_kwargs)
            register_action(
                # `name` kwarg will be reused for our module
                # registration, as well as for celery task registry
                _kwargs.get('name', func.__name__),
                tsk,
            )
            return tsk
        return __inner

    kwargs['bind'] = True
    kwargs['base'] = PipelineTask
    if len(args) == 1 and callable(args[0]):
        return wrapper(**kwargs)(args[0])
    return wrapper(*args, **kwargs)



class PipelineTask(Task):
    """Celery task wrapper, supports success/failure handler tasks
    and persisting source and parent return value into self.
    ^^ this is easier than relying on celery, since it forces
    task writers to write supporting boilerplate.
    """
    abstract = True

    def __call__(self, *args, **kwargs):
        # allow these exeptions to propagate, since if they werent
        # set it means the action was never prepared, which is a
        # violation of the api
        state = kwargs.pop('_pipeline_chain_state')

        workspace_instructions = kwargs.pop('_pipeline_workspace')

        # this needs to be run on the worker, not the broker
        self._pipeline_workspace = get_workspace(
            workspace_instructions['klass'],
            **workspace_instructions['kwargs']
        )

        self._pipeline_chain_state = state

        if not len(args) and not 'build_context' in self._pipeline_chain_state:
            raise ValueError('something went sideways')

        # store the result of the last-executed task (if one) into self.
        # we need to rely on the system to make sure that a task that
        # comes in with no args (but having a state) truly is the first task in a chain,
        # and it should already have a build context provided to it
        # by the executor.

        if len(args):  # maybe should be `len(args) > 1`, TODO?
            #TODO: i use type().__name__ here to reduce coupling
            # and avoid circular imports,  find a better way to do this.
            if type(args[0]).__name__ == 'BuildContext':
                # means we are a non-zeroth element in a chain,
                # and the pipeline task wraper has leveraged mutable
                # signatures to pass the build context to us.
                self._pipeline_chain_state['build_context'] = args[0]
                args = args[1:]

        #REMOVED, because who cares?
        # `source` should be the zeroth non-key arg
        #assert len(args) == 1, 'passing non key args to actions is not supported yet'

        return super(PipelineTask, self).__call__(*args, **kwargs)

    @property
    def build_context(self):
        """We dont want this to be used, but it' good for the transition to new api
        """
        if not hasattr(self, '_pipeline_chain_state'):
            raise AttributeError('task not executed in a pipeline')
        return self._pipeline_chain_state['build_context']

"""TODO: remove everything below this line.
This is the old API
"""
class Result(object):
    def __init__(self, result):
        self.result = result


class TaskResult(object):
    """Base container for a task result.

    Note: It is not required to use this class.
    As long as the thing added to context can be evaluated as a boolean,
    and it uses attribute access.

    Since we wish to chain arbitrary tasks, passing the return value
    of one task to the invocation of another, it is hepful for each
    task to share a common return signature.
    """
    def __init__(self, return_value, **kwargs):
        assert isinstance(return_value, bool)

        self.return_value = return_value
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        """This is required for celery 3.1, some internals
        use repr during serialization."""
        return "TaskResult({}, output='{}')".format(
            str(self.return_value),
            self.output
        )

    def __str__(self):
        return str(self.return_value)

    def __bool__(self):
        """Task return values should always implement this,'
        so a test for handler execution will work properly.
        """
        return self.return_value

#
# class TemplatedTaskMixin(object):
#     """Templated Task subclass.
#     Provides a ``render()`` method that a task can use to render
#     a predefined template using data from the provided source
#     (or elsewhere, I suppose).
#     # NOTE: celery Task class does not behave properly with respect
#     # to mro for __call__() method.  It is not possible to use
#     # multiple inheritance with chaining of __call__() for Task subclasses.
#     # ALL we can do is provide pure mixins, ust adding methods and not
#     # overriding or redefining them.
#     """
#     def __call__(self, retval=None, source=None, template=None, **kwargs):
#         # stick child task list into a private instance var
#         assert template
#         self._pipeline_render_context = {
#             'source': source,
#             'parent': retval
#         }
#         self._pipeline_template = template
#         #self._pipeline_render_retval = retval
#         #self._pipeline_render_source = source
#
#         return super(TemplatedTaskMixin, self).__call__(retval, source, **kwargs)
#     def render(self):
#         """Render template wth context."""
#         # render_context = {
#         #     'source': self.source,
#         #     'parent': self.parent_retval
#         # }
#         env = Environment()
#         env.globals.update(self._pipeline_render_context)
#         return env.from_string(self._pipeline_template).render()


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pipeline', '0006_auto_20150627_1934'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventhandler',
            name='criteria',
            field=jsonfield.fields.JSONField(null=True),
        ),
    ]



from pipeline.actions import action, TaskResult
from pipeline.tasks import TaskResult
from pipeline.command import CommandResult
import logging
logger = logging.getLogger(__name__)

class DT2Cop(object):

    def __init__(self, diff):
        self.diff = diff.decode('utf-8')
        self.new_classes = []
        self.new_methods = []

    def get_all_new_classes_from_file_diff(self, file_diff):
        """
        Calls get new methods
        """
        for i, line in enumerate(file_diff.split('+class')):
            if i == 0:
                continue
            else:
                new_class = '{}{}'.format('class', line)
                self.new_classes.append(new_class)
                self.get_all_new_methods_from_new_class(new_class)

    def get_all_new_methods_from_new_class(self, new_class):
        for i, method in enumerate(new_class.split('+    def')):
            if i == 0:
                continue
            new_method = '{}{}'.format('def', method)
            self.new_methods.append(new_method)

    def get_all_files_from_diff(self):
        # There is one index per file in the diff
        logger.critical(type(self.diff))
        files_diff = self.diff.split('index')
        for myfile in files_diff:
            self.get_all_new_classes_from_file_diff(myfile)

    def check_print_statements(self):
        """
        Diffs should not have any print statements.
        """
        # TODO use regex so method names that have print are not accounted for
        if 'print' in self.diff:
            msg = 'Print Violation. DT2COP id: {}'.format('BP13')
            return msg
        return ''

    def check_setup_teardown(self):
        """
        SetUp and TearDown methods must call super
        """
        for method in self.new_methods:
            if method.startswith('def setUp') or method.startswith('def tearDown'):
                if not 'super(' in method:
                    msg = 'Did not call super on test setUp() or tearDown(): DT2COP_ID: {}'
                    return msg.format('CT03')
        return ''

    def check_dunder_init(self):
        """
        Do not override dunder init if just calling super
        """
        for method in self.new_methods:
            if method.startswith('def __init__'):
                method = method.replace('+', '')
                method = method.split('\n')
                method = [x for x in method if x]  #filter(None, method)
                # Check if we are just overriding init
                if len(method) == 2:
                    if method[1].strip().startswith('super('):
                        msg = 'Do not override dunder init if just calling super. DT2COP_ID: {}'.format('CC01')
                        return msg
        return ''

    def check_manager_log_explicit_context(self):
        """
        self.log DOES NOT need an explicit context to be passed
        """
        for method in self.new_methods:
            if method.startswith('def DT20COP_lo'):
                method = method.replace('+', '')
                method = method.split('\n')
                method = [x for x in method if x]  #filter(None, method)
                for line in method:
                    if line.strip().startswith('self.log'):
                        if 'extra=self.context':
                            msg = 'self.log DOES NOT need an explicit context to be passed: DT2COP_ID: {}'.format('CT05')
                            return msg
        return ''

    def check_use_of_with_statement_with_get_manager(self):
        """
        Do not use with statement with getManager
        """
        for method in self.new_methods:
            if 'with getManager' in method:
                msg = 'Do not use with statement with getManager, DT2COP_ID: {}'.format('BP05')
                return msg
        return ''

    def check_use_of_dt_requests(self):
        """
        Use dt_requests module for all HTTP communication
        """
        if 'import requests' in self.diff:
            msg = 'Use dt_requests instead of requests. DT2COP_ID: {}'.format('BP08')
            return msg
        return ''

    def check_simplejson_import(self):
        if 'import json' in self.diff:
            msg = 'Use simplejson instead of json. DT2COP_ID: {}'.format('PL01')
            return msg
        return ''

    def run(self):
        output = []
        # Make sure we call `get_all_files_from_diff`
        self.get_all_files_from_diff()
        output.append(self.check_setup_teardown())
        output.append(self.check_print_statements())
        output.append(self.check_dunder_init())
        output.append(self.check_simplejson_import())
        output.append(self.check_use_of_dt_requests())
        output.append(self.check_use_of_with_statement_with_get_manager())
        output.append(self.check_manager_log_explicit_context())
        logger.critical(output)
        return '\n'.join(output)


@action(name='dtcop')
def dtcop(self, retval, source):
    """Run dtcop ona pull request"""
    output = DT2Cop(source.diff).run()
    return CommandResult(True, output=output)



"""
Simple event registration system.

"""
from cached_property import cached_property

from pipeline.executor import Executor
from pipeline.bases import Registry
from pipeline.eval import evaluate_criteria

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class Event(metaclass=Registry):
    """Base event class.
    Event lookup is expensive, this should be refactored.  Good enough for now.
    """
    #__abstract = True

    criteria = None

    def __init__(self, input_data):
        """Events should *only* require input_data to be deserialized.
        Filtering which events to fire should be done at the classmethod level.
        """
        self._input_data = input_data

    @classmethod
    def find_relevant(cls, input_data):
        """Find events relevant to a given input.
        Multiple events may match a single input.
        Matches the evewnts criteria against the event-specific
        deserialized input data.
        """
        available = cls.getlist()
        found = []
        for klass in available:
            instance = klass(input_data)
            if instance._is_relevant(): found.append(instance)
        return found

    @cached_property
    def data(self):
        """Return deserialized event data.
        """
        return {
            'source': self.source,
            'event': self.name
        }

    @cached_property
    def source(self):
        """Return the deserialized source."""
        raise NotImplementedError

    def _is_relevant(self):
        """Check to see if the event matches input, by
        matching self.criteria to self.data

        :returns: bool
        """
        return evaluate_criteria(self.data, self.criteria)

    @cached_property
    def name(self):
        """Provide a friendly name for consumers to indicate their
        interest in an event.
        TODO: this uses a workaround for private class variables
        (prepended with dunderscore).  See if there is a way that
        I can continue to use this technique (it is required for
        other parts of the system) but not have to resort to these
        getattrs to read them.
        """
        return getattr(self, '_{}__id'.format(self.__class__.__name__))


class EventHandler(object):
    """Class to act on an event."""
    def __init__(self, name, events=None, criteria=None, actions=None, enabled=True):
        """Default behavior:
            - events: handle no events
            - criteria: match all criteria
            - actions: do nothing
        """
        self.name = name
        self.events = events or []
        self.criteria = criteria
        self.actions = actions or []
        self.enabled = enabled

    def handle_event(self, event):
        """Execute the actions contained in the event handler
        using an executor, uisng the event data.
        """
        #z()
        if not self.enabled:
            #TODO should we communicate to caller using exception?
            return None

        try:
            source = event.data.pop('source')
        except KeyError:
            logger.error('Source is required in event data {}:{}'.format(self, event))
            raise

        executor = Executor()
        return executor.schedule(
            self.actions,
            source,
            **event.data
        )



from pipeline.actions import action, TaskResult
from pipeline.tasks import TaskResult
from pipeline.command import CommandResult
import logging

logger = logging.getLogger(__name__)


import requests
import json
import base64
import re

class Team(object):
    samurai_fighters = "Samurai Fighters"
    a_team = "A Team"
    usual_suspects = "Usual Suspects"
    djangsters = "Djangsters"
    rangers = "NY Rangers"
    islanders = "Islanders"
    scrum_bags = "Scrum Bags"
    think_kites = "Think Kites"
    falcons = "Falcons"
    param = "Param"
    killer_instinct = "Killer Instinct"
    core = "Core"
    others = "Others"
    fugu = "Fugu"
    cms = "CMS"

class GitUsers(object):

    #E001868
    #E001861

    user_dict = {
        "mike" : Team.samurai_fighters,
        "kulawansad" : Team.islanders,
        "yusuf-samiwala" : Team.others,
        "denzil-tarakan" : Team.usual_suspects,
        "E002070" : Team.scrum_bags, #"Macie"
        "veeravap" : Team.scrum_bags,
        "revanth-yesireddy" : Team.djangsters,
        "sushma-sharma" : Team.a_team,
        "ashish-gore" : Team.a_team,
        "DT2" : Team.others,
        "ananya-thakur":  Team.others,
        "patricia-byrne": Team.core,
        "andrew-keym" : Team.others,
        "phani-mallampalli" : Team.usual_suspects,
        "letor" : Team.islanders,
        "ganesh-adike" : Team.a_team,
        "madhava-challa" : Team.samurai_fighters,
        "Chaitanya" : Team.usual_suspects,
        "ravis" : Team.a_team,
        "jilesh-lakhani" : Team.others,
        "krishna-akkala" : Team.others,
        "praveen-pothana" : Team.others,
        "Miteshkumar-patel" : Team.think_kites,
        "alok-shelar" : Team.param,
        "amit-jain" : Team.think_kites,
        "kotlyav" : Team.islanders,
        "vikram-singam" : Team.samurai_fighters,
        "perminder-singh" : Team.islanders,
        "greg" : Team.djangsters,
        "E002327" : Team.others,
        "E001682" : Team.others,
        "rajesh-rana" : Team.others,
        "E002369" : Team.others,
        "Venkat-Vege" : Team.others,
        "nagarar" : Team.others,
        "kamlesh-vaishnav" : Team.others,
        "deepak-gahlot" : Team.others,
        "vijay-pahuja" : Team.others,
        "E002412" : Team.others,
        "E002411" : Team.others,
        "DRS" : Team.others,
        "akshay-vats" : Team.others,
        "anil-sharma" : Team.others,
        "Rajanikanth-Susarapu" : Team.scrum_bags,
        "Sean" : Team.usual_suspects,
        "ENVISAGE" : Team.others,
        "nader-abdelsadek" : Team.islanders,
        "john-o-shea" : Team.others,
        "E002015" : Team.others, #Shanta
        "E002515" :Team.others,
        "ravi-meesala" : Team.islanders,
        "Kapil-Bhatia" : Team.others,
        "shounak-de" : Team.core,
        "arun-chintam" : Team.core,
        "john-clements" : Team.core,
        "craig-bateman" : Team.core,
        "E002524" : Team.core,
        "Shanthosh-Stanislaus" : Team.others,
        "pavan-athota" : Team.others,
        "mike-miklancic" : Team.others,
        "Nicholas-Dunnaway" : Team.cms,
        "dan-heilman" : Team.others,
        "isakovn" : Team.others,
        "E002593" : Team.others,
        "Thomas-Kahnoski" : Team.others,
        "Willis-Godwin" : Team.others,
        "nick-maly" : Team.others,
        "patelm" : Team.others,
        "Charlee-Li"  : Team.others,
        "andrew-baird" : Team.others,
        "slava-filonenko"  : Team.others,
        "manish-kumar": Team.think_kites,
        "mikhail-starovoytov": Team.cms,
        "tommy-han": Team.islanders,
        "taras-kudla": Team.usual_suspects,
        "ajit-pai": Team.falcons,
        "miki725": Team.djangsters,
        "Dmitriy-Rozentul": Team.others,
        "Shreeyansh-Jain": Team.falcons,
        "khaled-porlin": Team.a_team,
        "sachin-jat": Team.think_kites,
        "Irene": Team.others,
        "naveen-reddy": Team.think_kites,
        "aniket-shelar": Team.think_kites,
        "agamdua": Team.samurai_fighters,
        "sundeep": Team.think_kites,
        "zaheer-shaik": Team.others,
        "kirt-gittens": Team.a_team,
        "milind": Team.djangsters,
        "Vimal-Jayachandran": Team.think_kites,
        "kumar-varadarajulu": Team.falcons,
        "thukaram-hulugappa": Team.cms,
        "Mark-Churney": Team.islanders,
        "scott-overholser": Team.others,
        "tony-mack": Team.falcons,
        "william-freeman": Team.fugu,
        "darrel-bonner": Team.others,
        "ram-chauhan": Team.others,
        "stephen-castle": Team.scrum_bags,
        "Abhishek-Sharma": Team.others,
        "ashok-vijay": Team.others,
        "Malarmannan-Polappan": Team.others,
        "timothy-chi": Team.others,
        "srinivas-thatipathy": Team.falcons,
        "arvind-singh": Team.others,
        "amninder": Team.scrum_bags,
        "miguel-molina": Team.others,
        "rajesh-kumar": Team.falcons,
        "Purandar-Palagala": Team.others,
    }

result = {}
def karma_rules(review_comment):
    """
    rules to increase or decrease score
    #blocker - minus
    :+1: - plus
    +1 - plus
    thank - plus
    #karma+
    #karma-
    #karma++
    #karma--
    """
    up_score_search_items = [(":+1:", "+1"), "thank"]
    down_score_search_items = ["#blocker"]

    score = 0
    analysis_dict = {}

    for item in up_score_search_items:
        if isinstance(item, tuple):
            for tuple_item in item:
                if tuple_item in review_comment:
                    analysis_dict[tuple_item] = analysis_dict.get(tuple_item, 0) + 1
                    score += 1
                    break
        else:
            if item in review_comment:
                analysis_dict[item] = analysis_dict.get(item, 0) + 1
                score += 1

    for item in down_score_search_items:
        if item in review_comment:
            analysis_dict[item] = analysis_dict.get(item, 0) + 1
            score -= 1

    karma_match = re.findall(r"#karma[\+\-]*", review_comment)

    for item in karma_match:
        for c in item:
            if c == "+":
                analysis_dict["#karma_plus"] = analysis_dict.get("#karma_minus", 0) + 1
                score += 1
            elif c == "-":
                analysis_dict["#karma_minus"] = analysis_dict.get("#karma_minus", 0) + 1
                score -= 1

    return score, analysis_dict


def update_result(team, new_score, analysis_dict, result):

    result[team]["score"] = new_score

    if "analysis" not in result[team]:
        result[team]["analysis"] = {}

    a_dict = result[team]["analysis"]
    for k, v in analysis_dict.iteritems():
        value = a_dict.get(k, 0) + v
        result[team]["analysis"][k] = value


@action(name='karmabot')
def karmabot(self, retval, source):
    if type(source) == 'PrBuilderStatusSource':
        return TaskResult(True)  #hack
    from pipeline.command import CommandResult
    return CommandResult(True, output=karma_rules(source.body))




""" Pipeline Executors.

They accept a list of actions, and execute them in some order using the
provided Source and a BuildContext.

"""
from celery import chain

from pipeline.context import BuildContext

import logging
logger = logging.getLogger(__name__)


class Executor(object):
    """Task scheduler.  Currently using a single chain,
    but a future impl will use more of the celery canvas (perhaps allowing the
    caller to specify, or using a dependency graph to use the most efficient
    canvas)."""
    @staticmethod
    def schedule(actions, source, **kwargs):
        """Execute actions using source, return the AsyncResult of the last task."""
        tasks = []
        for idx, action in enumerate(actions):
            if idx == 0:
                # Only pass in context for the first task in the chain,
                # as celery mutable signatures will take care
                # of propagating it to the other tasks.
                build_context = BuildContext(**kwargs)
                partial = action.prepare(source, build_context)
            else:
                partial = action.prepare(source)

            tasks.append(partial)  # action.partial)

        canvas = chain(*tasks)

        logger.debug('got {} actions {}'.format(len(actions), actions))
        logger.debug('chaining tasks {}'.format(tasks))

        ret = canvas.apply_async(retry=False)  #FIXME retry policy

        return ret




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pipeline', '0005_rename_source_class'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='configuration',
            name='source',
        ),
        migrations.RemoveField(
            model_name='configurationsource',
            name='source_class',
        ),
        migrations.AddField(
            model_name='configuration',
            name='criteria',
            field=jsonfield.fields.JSONField(default=[]),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='configuration',
            name='events',
            field=jsonfield.fields.JSONField(default=[]),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='configurationsource',
            name='event',
            field=models.CharField(max_length=255, default=[]),
            preserve_default=False,
        ),
    ]



import os
import json
import subprocess

from pipeline.bases import Registry
from pipeline.workspace import Workspace
from pipeline.actions import action, TaskResult


import logging
logger = logging.getLogger(__name__)


@action
def shell_command(self, source, commands):
    """Run some shell commands.
    Holy shit this is easy.
    """
    assert isinstance(commands, (list, tuple))

    with self._pipeline_workspace as workspace:
        workspace.acquire_source(source)
        for command in commands:
            try:
                workspace.session.check_call(command)
            except subprocess.CalledProcessError as ex:
                logger.critical(str(ex))
                break

        return workspace.session.log


class BaseOutputFormatter(object):
    def __call__(self, output):
        return output.strip()


class LineOutputFormatter(BaseOutputFormatter):
    """The interface of this formatter is that it should accept
    unicode input, and return a list of strings.
    """
    def __call__(self, output):
        """Return list-ified output."""
        return super(LineOutputFormatter, self).__call__(output).split(os.linesep)


class CommandResult(TaskResult):
    """Subclass of TaskResult that understands that shell commands
    return 0 for success."""
    def __init__(self, return_value, output=None):
        #self.return_value = bool(return_value==0)
        self.returncode = return_value
        super(CommandResult, self).__init__(bool(return_value==0), output=str(output))
    #def __bool__(self):
    #    return self.return_value == 0


class BakedCommand(metaclass=Registry):
    """Base class for shell commands."""
    __id = None  # required
    install = None  # optional
    # removed until I figure out how to transform input
    formatter = BaseOutputFormatter

    def __init__(self, context):
        #self.source = source
        #self.retval = retval
        self.context = context

    def instructions(self):
        """Command entry point.
        :param source: required
        :returns list of subprocess-friendly lists
            ex:
                return [
                    ['ls', '-l'],
                    ['uname', '-a']
                ]
        """
        assert hasattr(self, 'context')
        raise NotImplementedError


class CommandError(Exception):
    pass


def baked_command_task(self, context, **kwargs):
    """This is the template function for a task that will run
    a shell command inside of a ``pipeline.workspace.workspace``.

    This wont work by itself, needs to be wrapped with shared_task
    (which is accomplished by calling ``register_command_task``).

    TODO: fix error handling inthis module, defer command errors to workspace.py?

    """
    assert hasattr(self, 'workspace_class')
    assert hasattr(self, 'command_name')

    workspace_cls = Workspace.get(self.workspace_class)
    with workspace_cls(name='command-{}'.format(self.command_name)) as wkspc:

        # we pass source everydamnwhere.  0xbadidea?
        cmd = BakedCommand.factory(self.command_name, context=context)

        #TODO: use check_call here so I can catch CalledProcessError
        # instead of these nested if stmts.
        result = wkspc.acquire_source(context.source)

        if result == 0:
            if cmd.install:
                install_ret = wkspc.run_command(cmd.install)
                if not install_ret == 0:
                    raise CommandError("error {} running {} as {} with {}".format(
                        install_ret, cmd.install, wkspc.user, json.dumps(wkspc.environ, indent=2)
                    ))

            #TODO: support multiple insturctions
            output = []
            assert len(cmd.instructions())
            for instruction in cmd.instructions():
                #instruction = cmd.instructions()[0]  # just grab the first
                _output, errcode = wkspc.get_output(instruction)
                output.append(cmd.formatter()(_output))
                logger.debug('command {} returned {}'.format(instruction, errcode))
            output = '\n'.join(output)
        else:
            #TODO: return output?
            raise CommandError("error {} running {} as {} with {}".format(
                result, context.source.acquisition_instructions, wkspc.user, json.dumps(wkspc.environ, indent=2)
            ))

        return TaskResult(errcode==0, errcode=errcode, output=output)


def register_baked_command_task(command_name, workspace_class='workspace', task_func=None, **kwargs):
    """Generate and register a TaskAction that wraps a shell command.

    :param command_name: the friendly task name that will
        be stored in the registry. Also will be used as celery
        internal task name.
    :param workspace_class: friendly/registered name of a subclass of
        ``pipeline.workspace.workspace``.  This will be used to wrap the
        execution of the command on a worker.
    :param task_func: a template function to be used to create the
    command task.  Primarily used for testing.
    :param kwargs: passed into ``shared_task`` to help define the task object.
    :returns: ``celery.local.Proxy`` instance - thsi represents the actual
        task stored inside the celery app.
    """
    if not task_func:
        task_func = baked_command_task

    assert hasattr(task_func, '__call__')

    return action(
        task_func,
        name=command_name,
        command_name=command_name,
        workspace_class=workspace_class,
        **kwargs
    )




import re

from pipeline.command import (
    BakedCommand,
    register_baked_command_task,
)


class CoverageDiffCommand(BakedCommand):
    """
    Command subclass for installing and running flake8-diff command.
    """
    __id = 'coverage_diff'
    install = ['pip', 'install', '--upgrade', 'diff_cover']

    def instructions(self):
        """Return subprocess cmd for execution.
        :returns: subprocess-friendly list to run diff_cover

        Expects self.source to have an 'output' that will contain coverage xml.
        Quack like a duck.
        """
        return [
            'http_proxy= curl {} --user {}:{} -o  coverage.xml'.format(
                self.retval.url,
                self.retval.user,
                self.retval.passwd
            ),
            [
            'diff-cover',
            'coverage.xml',
            '--compare-branch=origin/{}'.format(self.source.dest_branch),
            ]
        ]


register_baked_command_task('coverage_diff', workspace_class='python_workspace')




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pipeline', '0005_auto_20150627_1932'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventHandler',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('events', jsonfield.fields.JSONField(default=[])),
                ('criteria', jsonfield.fields.JSONField(default=[])),
                ('actions', jsonfield.fields.JSONField()),
                ('app_name', models.CharField(max_length=64, blank=True)),
            ],
        ),
        migrations.DeleteModel(
            name='Configuration',
        ),
    ]




from pipeline.command import (
    BakedCommand,
    register_baked_command_task,
)




class ProspectorCommand(BakedCommand):
    """
    Command subclass for installing and running prospector command.
    """
    __id = 'prospector'
    install = ['pip2', 'install', '--upgrade', 'prospector']


    def instructions(self):

        return [[
            'prospector',
        ]]


register_baked_command_task('prospector', workspace_class='python_workspace')


"""
Pipeline tasks and task helpers.

Place all code that interacts with celery shared_task or app.task in this module;
these will have side-effects, and it's good to keep them in a single module so that errant
utility imports don't trigger these side effects.
"""
import celery
from celery.canvas import Signature
import six

"""
Code below this point is fully functional, and is the result of a large amount of
analysis, reading celery source code, and testing.  But, pipeline does not need it.
Dammit.
"""
@celery.task(bind=True)
def task_predicate_match(self, *args, **kwargs):
    """REQUIRES Celery 3.2

    Schedule one of two tasks, based on evaluation of a predicate
    using the result of a parent task as context.
    If this is created as an immutable signature, this will not function.
    Discards the current task.

    :param *args: 2- or 3-tuple
        1. result of previous task (if this task is not immutable)
        2. task signature to execute on success
        3. task signature to execute on failure
    :param kwargs['predicate']: expression to eval to determine success
    :return: None
        Callers should not expect the return value of this
        task to be useful/usable, since it is being replaced.

    It is important to note that this task does not return a value,
    so any signature that follows it in a chain will not receive a result,
    and if this signature is the last in a chain it will not return
    any result to the original caller.
    """
    assert len(args) in (2,3)

    idx = 0
    if not isinstance(args[0], Signature):
        # signature for the current task is not immutable, so we
        # should assume that the first arg is the result of the
        # previous task in a chain.
        result = args[0]
        idx += 1

    # ^^ note that `result` is not part of local scope if we are
    # immutable; since it's only used for the predicate, we will allow
    # the eval chec to raise an exception, which will fail this task.

    success = args[idx]
    failure = args[idx+1]

    # default behavior is to evaluate `result` for truthy-ness
    predicate = kwargs.get('predicate', "result")


    assert not success or isinstance(success, Signature)
    assert not failure or isinstance(failure, Signature)
    assert success or failure
    assert isinstance(predicate, six.string_types)

    try:
        predicate_met = eval(predicate)
    except Exception as exc:
        # you may not recv a report of this exception
        # task_logger.error('cannot evaluate predicate {} ({})'.format(
        #     predicate,
        #     str(exc))
        # )
        raise
    else:
        if predicate_met and success:
            # task_logger.error('replacing task {} with {}'.format(
            #     self.name, success.task
            # ))
            self.replace(success)
        elif failure:
            # task_logger.error('replacing task {} with {}'.format(
            #     self.name, failure.task
            # ))
            self.replace(failure)


def exclusive_group(success, failure, **kwargs):
    """Conditionally schedule one or another task
    based on some predicate.

    :param success: celery signature/canvas element
    :param failure: celery signature/canvas element
    :param kwargs['predicate']: expression
    """
    # we send two params to task, it gets 3.  Magic!
    return task_predicate_match.s(success, failure, **kwargs)


def conditional_task(task, **kwargs):
    """Conditionally schedule a task based on a
    given predicate.

    :param task: celery signature/canvas element
    :param kwargs['predicate']: expression
    """
    return task_predicate_match.s(task, None, **kwargs)





"""
Pipeline

This package (ab)uses the registry pattern, so it it is critical that
modules using registries are imported as early as possible
"""
__version__ = '0.6'


import pipeline.signals
import pipeline.workspace


#TODO(mikew): Move these from explicit CommndTask subclasses to something
#TODO: more generic, so we don't require python code to implement
#TODO: either of them (or any other shell commands)
#from pipeline.modules import flake8, prospector


from pipeline.modules import debug
#from pipeline.modules import dtcop
#from pipeline.modules import jenkins
#from pipeline.modules import coverage
#from pipeline.modules import karma





"""
``pipeline.context``

Build Context

TODO:
    - Move BuildContext results from simple dicts to objects,
    for easy attribute access.
    - Task name is not going to be sufficient for ``update``,
    because the same task may be executed multiple times in a chain.
    Need to somehow apply ``name`` param of action as it was provided
    by the user.
    Ex:
        { "task": "some_class", "name": "my_task_exec_thingy"}
"""
import six
from jinja2 import Environment

from pipeline.eval import safe_eval


class BuildContext(object):
    """Container for build-time data.
    Stores ``pipeline.bases.Source`` object, as well as
    the result for tasks executed in a chain.

    No need to worry about mutable state here, since this
    object will be serialized and deserialized by celery
    between execution of tasks.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

        #TODO this should be some magic in getattr
        for k, v in self.kwargs.items():
            setattr(self, k, v)

        #TODO: read these in getattr
        self.results = {}

    def update(self, action_or_name, result):
        """Update the build context state with the result
        of a task.
        FIXME action_or_name
        """

        if type(action_or_name) == 'TaskAction':
            name = action_or_name.__name__
        else:
            name = action_or_name

        assert not name in self.results.keys(), \
            'multiple task exec not yet supported'

        self.results[name] = result

        return self

    def render(self, template, **kwargs):
        """Render a template using build context data.
        """
        env = Environment()
        env.globals.update(self.eval_context)
        env.globals.update(kwargs)
        return env.from_string(template).render()

    def render_params(self, source, *args, **kwargs):
        """Render all the things!
        Apply rendering to all the args and kwargs that were
        provided to a task.
        Need a better solution to this brute-force copy/paste job.
        """
        for idx, arg in enumerate(args[2:]):
            if isinstance(arg, six.string_types):
                args[idx] = self.render(arg, source=source)
            elif isinstance(arg, (list, tuple)):
                dest = []
                for item in arg:
                    dest.append(self.render(item, source=source))
                args[idx] = dest

        for k, v in kwargs.items():
            if isinstance(v, six.string_types):
                kwargs[k] = self.render(v, source=source)
            elif isinstance(v, (list, tuple)):
                dest = []
                for item in v:
                    dest.append(self.render(item, source=source))
                kwargs[k] = dest

        return args, kwargs

    def evaluate(self, expression):
        """Evaluate an expression using context.
        """
        return safe_eval(expression, self.eval_context)

    @property
    def eval_context(self):
        """Place interesting things in a dict to be sent to eval.
        """
        context = {}

        for k, v in self.results.items():
            context.update({k: v})
        for k, v in self.kwargs.items():
            context.update({k: v})
        return context


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from pip.req import parse_requirements

install_reqs = parse_requirements(
    'requirements.txt',
    session=False
)
reqs = [str(ir.req) for ir in install_reqs]

# see above
#version = pipeline.__version__
version = '0.7'

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

#readme = open('README.rst').read()
#history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='celery-pipeline',
    version=version,
    description="""Runtime-configured execution pipeline built on celery.""",
    author='Mike Waters',
    author_email='robert.waters@gmail.com',
    url='https://github.com/mikewaters/pipeline',
    packages=[
        'pipeline',
    ],
    include_package_data=True,
    install_requires=reqs,
    license="BSD",
    zip_safe=False,
    keywords='pipeline',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
    ],
)



"""
Flake8Diff support

"""
import re

from pipeline.command import (
    BakedCommand,
    register_baked_command_task,
    LineOutputFormatter
)


class Flake8LintFormatter(LineOutputFormatter):
    """Format linter errors.
    Currently unused, fix is part of Milestone2
    (violation comment on individual commits instead of on pull request)
    """
    def __call__(self, output):
        """Return list-ified output."""
        file_re = re.compile(r'^Found violations: (?P<path>.*)$')
        violations = []
        _file = None
        for line in super(Flake8LintFormatter, self).__call__(output):
            match = file_re.match(line)
            if match:
                _file = match.group('path')
                continue
            elif line.startswith('\t'):
                violations.append((_file, line.strip()))
                continue

        return violations


class Flake8DiffCommand(BakedCommand):
    """
    Command subclass for installing and running flake8-diff command.
    """
    __id = 'flake8_diff'
    install = ['pip', 'install', '--upgrade', 'flake8-diff']
    # removed - will be resolved with "allow output transformation for actions"
    # formatter = Flake8LintFormatter

    def instructions(self):
        """Return subprocess cmd for execution.
        :returns: subprocess-friendly list to run flake8-diff

        Expects self.source to be a ``github3.pulls.pull_request`` instance.
        Quack like a duck.
        """
        return [[
            'flake8-diff',
            # will always be origin, since we initially cloned `base`
            'origin/{}'.format(self.context.source.dest_branch),
            #TODO implement configurable params for shell commands
            '--flake8-options', '--ignore=E501', '--exclude=\*_tables.py',
        ]]


register_baked_command_task('flake8_diff', workspace_class='python_workspace')




"""
Evaluation helpers.


"""
import re

ALLOWED_BUILTINS = ('bool',)


def get_default_builtins():
    """Get the allowed eval() builtins.
    """
    ret = {}
    for k, v in __builtins__.items():
        if k in ALLOWED_BUILTINS:
            ret[k] = v
    return ret

def safe_eval(expression, _locals, _globals=None):
    """Run eval in a semi-safe manner.

    Current impl just runs eval().  Future iterations will attempt to
    do this in a safer manner, using ast.
    see: http://code.activestate.com/recipes/364469-safe-eval/
    """
    if not _globals:
         _globals = {'__builtins__': get_default_builtins()}

    _locals = _locals or {}

    # `expression` might just be a string, so catch some exceptions here
    # and just return expression.
    try:
        ret = eval(
            expression,
            _globals,
            _locals
        )
    except NameError:
        return expression
    except Exception as ex:
        print(ex)
        return expression
    return ret



def evaluate_single_criterion(data, criterion):
    """Determine if a one of criteria match the event data.
    This is a very, very trivial implementation of something
    that could be interesting, if given time to hack on it.

    An example of criteria is:
        ['object.attribute', 'is', 'some value']

    :param data: dict of event, source
    :param criterion: 3-tuple of (lvalue, operator, rvalue)
    :returns boolean
    """
    oper = criterion[1]
    try:
        lvalue = safe_eval(criterion[0], data)
    except:
        #TODO
        raise
    rvalue = criterion[2]

    if oper == 'is':
        return lvalue == rvalue
    elif oper == 'in':
        return lvalue in rvalue
    elif oper == 'like':
        #TODO re.flags?
        return bool(re.search(
            rvalue, lvalue
        ))
    elif oper == 'not':
        return lvalue != rvalue
    elif oper == 'not in':
        return lvalue not in rvalue
    elif oper == 'not like':
        #TODO re.flags?
        return not bool(re.search(
            rvalue, lvalue
        ))
    else:
        raise NotImplementedError(
            'operator {} not supported'.format(oper)
        )

def evaluate_criteria(data, criteria):
    """Check to see if the criteria matches data.

    - if criteria is None, it should always match
    - if criteria is [], it should never match
    - otherwise, evaluate against deserialized incoming data

    :returns: bool
    """
    if criteria is None:
        return True

    assert isinstance(criteria, (list, tuple))
    if not len(criteria):
        return False

    for criterion in criteria:
        if not evaluate_single_criterion(data, criterion):
            return False

    return True